# version 09/25/08
#######################################################################
##                                                                   ##
##  textbook.g                       William Paulsen                 ## 
##                                                                   ## 
##  This file contains the GAP functions that are used for the text  ##
##  "Abstract Algebra - An Interactive Approach" by William Paulsen  ##
##                                                                   ##
##  Many functions use the library "newrings.g" by Alexander Hulpke  ##
## If it has not been loaded yet, load it first.                     ##
##                                                                   ##
#######################################################################

if IsBound(RingByStructureConstants) = false then 
   Read("c:/gap/newrings.g"); # should also be in the gap directory.
fi;


#######################################################################
##                                                                   ##
##  Set ITER_POLY_WARN  to false, so that we can define polynomials  ##
##  over a new ring without a warning message.                       ##
##                                                                   ##
#######################################################################

ITER_POLY_WARN := false;

#############################################################################
## 
## InitTerry()
## 
## InitTerry sets up the first group to explore.  Since students do  
## not yet know how to form a group in GAP, this does the work for   
## them.                                                             
#############################################################################

InitTerry := function()
local s;
  # Because we want the 6 variables to be defined after the function is executed, 
  # this first defines s as a long string that would set up the group, and then
  # the string is read as input.  This allows us to make InitTerry a function,
  # as opposed to a file that is read.
  s := "ff := FreeGroup(\"Stay\",\"RotLft\",\"RotRt\",\"Spin\",\"FlipLft\",\"FlipRt\");\n";
  # Define a group with 6 generators, and make every element of the group a generator.
  # Obviously inefficient, but it sets up the first group.
  s := Concatenation(s,"gg := ff/[ff.1,ff.2^3,ff.3^3,ff.2*ff.3,ff.4^2,ff.5^2,");
  s := Concatenation(s,"ff.6^2,ff.6*ff.3*ff.5,ff.4*ff.3*ff.6];\n");
  s := Concatenation(s,"Stay := gg.1;\n");
  s := Concatenation(s,"RotLft := gg.2;\n");
  s := Concatenation(s,"RotRt := gg.3;\n");
  s := Concatenation(s,"Spin := gg.4;\n");
  s := Concatenation(s,"FlipLft := gg.5;\n");
  s := Concatenation(s,"FlipRt := gg.6;\n");
  s := Concatenation(s,"Terry := [Stay, FlipRt, RotRt, FlipLft, RotLft, Spin];\n");
  s := Concatenation(s,"Print(Terry,\"\\n\");\n");
  Read(InputTextString(s));
end;

#############################################################################
##
##  MultTable( <group or ring> )
##
##  This displays the multiplication table of the group or ring, or list of 
##  elements from a group or ring.  If 0 is included in the set of integers, 
##  then the operation is addition modulo (1 + largest number).  If 0 is not
##  in the list of integers, then the operation is multiplication modulo 
##  (1 + largest number).  Otherwise, the elements are not integers, but elements
##  of a previously defined group.  We can also have a list of cosets, in which
##  case the multiplication table of the quotient group is displayed.
##  There are many options that can be set for this function.
##
#############################################################################

# Options for the MultTable command
MultTableHeader := true;
DisplayIdentity := "e ";
NumberElements := false;
FilterCommas := false;
IntPermMultiplication := false;
QuotientRing := false;

# Allow a way for these options to be reset back the way they were.
ResetTableOptions := function() 
  MultTableHeader := true;
  DisplayIdentity := "e ";
  NumberElements := false;
  FilterCommas := false;
  IntPermMultiplication := false;
  QuotientRing := false;
end;

# These functions will be introduced later, but to prevent 
# "Variable must have a value" errors, we set them to 0 here
Mult := 0;
AddC := 0;
PermToInt := 0;

MultTable := function(L)
   local i, j, k, LL, SizeG, TT, G, Table, base, m, s, p, biggest, filter, spacing;
   filter := function(obj)
      # this local function converts obj to a string, then strips the spaces from the 
      # string, and replaces <identity ...> with the string in DisplayIdentity.
      local obj2, ii, jj, kk, ss, out;
      # ss := String(obj);
      # Unfortunately, String(obj) does not always work, so here is a workaround
      obj2 := obj;
      if IntPermMultiplication then 
         obj2 := PermToInt(obj);
      fi;
      ss := "";
      out := OutputTextString(ss,false);
      PrintTo(out,obj2);
      RemoveCharacters(ss," ");
      if (FilterCommas = true) then 
         RemoveCharacters(ss,",");
      fi;
      ii := Position(ss,'<');
      if ii <> fail then
         # since we found a <, we can assume that everything up to > will be converted to
         # the identity element.
         jj := Position(ss,'>',ii);
         if jj <> fail then
            # here is another twist.  When the identity element is displayed in a pc
            # group, it is <identity> of ...  (of... is not in the brackets!)
            if Length(ss) >= jj + 5 then
               if ss{[jj+1..jj+5]} = "of..." then
                  jj := jj + 5;
               fi;
            fi;
            kk := ss{[ii..jj]};
            ss := ReplacedString(ss,kk,DisplayIdentity);
         fi;
      fi;
      return ss;
   end;
   LL := List(L);
   # Only in the case of permutation do we reorder the elements according to the 
   # PermToInt function
   if (IntPermMultiplication and PermToInt((1,2)) > 1) then
      Sort(LL, function(v,w) return PermToInt(v) < PermToInt(w); end );
   fi;
   # remove duplicates from the list
   for i in Reversed([2..Size(LL)]) do
      for j in [1..(i-1)] do
         if (LL[i] = LL[j]) then 
             Remove(LL,i);
             break;
         fi;
      od;
   od;
   G := Flat(LL); # find the base group
   SizeG := Size(G);
   # base = 0 means G is a group, base > 0 means multiply mod base, base < 0 means add mod base.
   base := 0;
   if (Number(G,IsInt) = SizeG) then
      base := Maximum(G) + 1;
      if (0 in G) then
         base := -base;
      fi;
   fi;
   # Find the element that takes the most characters to print
   biggest := 0;
   for i in [1..Size(LL)] do
      s := filter(LL[i]);
      if Size(s) > biggest then
         biggest := Size(s);
      fi;
   od;
   spacing := "  ";
   # Change spacing to " " is space is tight.
   if (biggest+2)*(Size(LL)+1)-3 > 80 then 
      spacing := " ";
   fi;   
   if NumberElements then 
      spacing := " ";
   fi;
   Print("\n");
   if (MultTableHeader or NumberElements) then
      if base < 0 then
         Print(String("+",-biggest));
      else 
         Print(String("*",-biggest));
      fi;
      Print("|");
      for i in [1..Size(LL)] do
         if NumberElements then
            Print(String(i,-2));
         else
            Print(String(filter(LL[i]), -biggest));
         fi;
         if i < Size(LL) then
            Print(spacing);
         fi;
      od;
      Print("\n");
      for i in [1..biggest] do
         Print("-");
      od;
      Print("+");
      m := Size(LL)*biggest  + (Size(LL) - 1)*Length(spacing);
      if NumberElements then 
         m := Size(LL)*3  - 1;
      fi;
      for i in [1..m] do
         Print("-");
      od;
      Print("\n");
   fi;
   for i in [1..Size(LL)] do
      if (MultTableHeader or NumberElements) then
         s := String(filter(LL[i]), -biggest);
         Print(s,"|");
      fi;
      for j in [1..Size(LL)] do  
         # find the product p 
         if IsList(LL[i]) then
            p := Mult(G,LL[i],LL[j]); # Multiply two cosets
            m := "";
         else
            m := "";
            if (base > 0) then
               p := (LL[i] * LL[j]) mod base;
            elif (base < 0) then
               p := (LL[i] + LL[j]) mod (- base);
            else 
               p := LL[i]*LL[j];
            fi;
         fi;
         for k in [1..Size(LL)] do
            if (IsList(p) and IsList(LL[k])) then
               if IsSubset(LL[k],p) then
                  if QuotientRing then
                     m := LL[k];
                  else   
                     m := p;
                  fi;
               fi;
            fi;
            if (p = LL[k]) then
               m := LL[k];
               break;
            fi;
         od;
         if p = 0 then
            m := 0;
         fi;
         if NumberElements then
            Print(String(k,-2));
         else
            Print(String(filter(m), -biggest));
         fi;
         if j < Size(LL) then
            Print(spacing);
         fi;
      od;
      Print("\n");
   od;
   Print("\n");
 end;

#############################################################################
##
##  AddTable( <ring> )
##  
##  Does the same thing as MultTable for a ring, but displays the addition
##  table.
##
#############################################################################

AddTable := function(L)
   local i, j, k, LL, SizeG, TT, G, Table, m, s, p, biggest, filter, spacing;
   filter := function(obj)
      # this local function converts obj to a string, then strips the spaces from the 
      # string, and replaces <identity ...> with the string in DisplayIdentity.
      local obj2, ii, jj, kk, ss, out;
      # ss := String(obj);
      # Unfortunately, String(obj) does not always work, so here is a workaround
      obj2 := obj;
      ss := "";
      out := OutputTextString(ss,false);
      PrintTo(out,obj2);
      RemoveCharacters(ss," ");
      if (FilterCommas = true) then 
         RemoveCharacters(ss,",");
      fi;
      # Anything in < > gets replaced with the string DisplayIdentity
      ii := Position(ss,'<');
      if ii <> fail then
         jj := Position(ss,'>',ii);
         if jj <> fail then
            # here is another twist.  When the identity element is displayed in a pc
            # group, it is <identity> of ...  (of... is not in the brackets!)
            if Length(ss) >= jj + 5 then
               if ss{[jj+1..jj+5]} = "of..." then
                  jj := jj + 5;
               fi;
            fi;
            kk := ss{[ii..jj]};
            ss := ReplacedString(ss,kk,DisplayIdentity);
         fi;
      fi;
      return ss;
   end;
   LL := List(L);
   # Only in the case of permutation do we reorder the elements according to the 
   # PermToInt function
   if (IntPermMultiplication and PermToInt((1,2)) > 1) then
      Sort(LL, function(v,w) return PermToInt(v) < PermToInt(w); end );
   fi;
   # remove duplicates from the list
   for i in Reversed([2..Size(LL)]) do
      for j in [1..(i-1)] do
         if (LL[i] = LL[j]) then 
             Remove(LL,i);
             break;
         fi;
      od;
   od;
   G := Flat(LL); # find the base group or ring
   SizeG := Size(G);
   # Find the element that takes the most characters to print
   biggest := 0;
   for i in [1..Size(LL)] do
      s := filter(LL[i]);
      if Size(s) > biggest then
         biggest := Size(s);
      fi;
   od;
   spacing := "  ";
   # Change spacing to " " is space is tight.
   if (biggest+2)*(Size(LL)+1)-3 > 80 then 
      spacing := " ";
   fi;   
   if NumberElements then 
      spacing := " ";
   fi;
   Print("\n");
   if (MultTableHeader or NumberElements) then
      Print(String("+",-biggest));
      Print("|");
      for i in [1..Size(LL)] do
         if NumberElements then
            Print(String(i,-2));
         else
            Print(String(filter(LL[i]), -biggest));
         fi;
         if i < Size(LL) then
            Print(spacing);
         fi;
      od;
      Print("\n");
      for i in [1..biggest] do
         Print("-");
      od;
      Print("+");
      m := Size(LL)*biggest  + (Size(LL) - 1)*Length(spacing);
      if NumberElements then 
         m := Size(LL)*3  - 1;
      fi;
      for i in [1..m] do
         Print("-");
      od;
      Print("\n");
   fi;
   for i in [1..Size(LL)] do
      if (MultTableHeader or NumberElements) then
         s := String(filter(LL[i]), -biggest);
         Print(s,"|");
      fi;
      for j in [1..Size(LL)] do  
         # find the product p 
         if IsList(LL[i]) then
            p := AddC(G,LL[i],LL[j]); # Add two coset
            m := ""; #if not in group, display blank
         else
            m := ""; #if not in group, display blank
            p := LL[i]+LL[j];
         fi;
         for k in [1..Size(LL)] do
            if (p = LL[k]) then
               m := LL[k];
               break;
            fi;
         od;
         if p = 0 then
            m := 0;
         fi;
         if NumberElements then
            Print(String(k,-2));
         else
            Print(String(filter(m), -biggest));
         fi;
         if j < Size(LL) then
            Print(spacing);
         fi;
      od;
      Print("\n");
   od;
   Print("\n");
 end;
   
#############################################################################
##
##  ListGroup( <group> )
##
##  For a group defined by generators, the way that Mathematica defines the group,
##  this gives a list of elements in the format and order that Mathematica
##  uses.  That is, each element is a product of generators in "alphabetical order".
##
#############################################################################

ListGroup := function(g)
   local L, Ord, n, iList, pointer, Cont, GList, ele;
   if (IsGroup(g) = false) then
      Print("Not a group\n");
      return fail;
   fi; 
   if (Size(g) = infinity) then
      Print("Infinite group\n");
      return fail;
   fi;
   L := GeneratorsOfGroup(g);
   Ord := List(L, x->Order(x));
   # do a arbitrarily nested for loop
   n := Size(L);
   GList := [ ];
   iList := List([1..n],x->0); # Vector of zeros, of length n
   Cont := true;
   while Cont do
      ele := Product([1..n],x->L[x]^iList[x]);
      if not (ele in GList) then
         Add(GList,ele);
      fi;
      # increment the iList
      pointer := 1;
      iList[pointer] := iList[pointer] + 1;
      while (iList[pointer] >= Ord[pointer]) do
         iList[pointer] := 0;
         pointer := pointer + 1;
         if (pointer > n) then
            Cont := false;
            pointer := 1;
         fi;
         iList[pointer] := iList[pointer] + 1;
      od;
   od;
   return GList;
end;

#############################################################################
##
##  Mult( <group or list of cosets>, <subset1>, <subset2> )
##
##  This multiplies the two subsets together to form a new subset.  
##  With this we can multiply two elements of the quotient group. 
##  We can even form the quotient group of the quotient group, for
##  the second isomorphism theorem.  The subsets can be single elements of
##  the group.
##
#############################################################################

Mult := function( G, H, Y )
   local GG, XX, YY, i, j, FlatG, SizeG, p, base, k, L;
   # Find the base group
   GG := List(G);
   FlatG := Flat(GG);
   # Determine the operation
   SizeG := Size(FlatG);
   # base = 0 means G is a group, base > 0 means multiply mod base, 
   # base < 0 means add mod base.
   base := 0;
   if (Number(FlatG, IsInt) = SizeG) then
      base := Maximum(FlatG) + 1;
      if (0 in FlatG) then
         base := -base;
      fi;
   fi;
   L := [];
   if (IsList(H) or IsGroup(H) or IsRing(H)) then
      XX := List(H);
   else
      XX := [H];
   fi;
   if (IsList(Y) or IsGroup(Y) or IsRing(Y)) then
      YY := List(Y);
   else
      YY := [Y];
   fi;
   # assume first that XX and YY are simple lists.
   for i in [1..Size(XX)] do
      for j in [1..Size(YY)] do
         if(IsList(XX[i]) or IsList(YY[j]) ) then
            Add(L, Mult(FlatG, XX[i], YY[j])); # recursively defined for sets of sets
         else
            # find the product p
            if (base > 0) then
               p := (XX[i] * YY[j]) mod base;
            elif (base < 0) then
               p := (XX[i] + YY[j]) mod (- base);
            else
               p := XX[i]*YY[j];
            fi;
            for k in [1..Size(GG)] do
               if (p = GG[k]) then
                  Add(L, GG[k]);
                  break;
               fi;
            od;
         fi;
      od;
   od;
   Sort(L);
   # remove duplicates from the list
   for i in Reversed([2..Size(L)]) do
      for j in Reversed([1..i-1]) do
         if (L[i] = L[j]) then
             Remove(L,i);
             break;
         fi;
      od;
   od;
   return L;
end;

#############################################################################
##
##  AddC( <ring or list of cosets>, <subset1>, <subset2> )
##
##  This does the same as Mult does for groups, except it uses the additive
##  structure of the ring instead.  Hence, we can add two ring cosets.
##
#############################################################################

AddC := function( G, H, Y )
   local GG, XX, YY, i, j, FlatG, SizeG, p, base, k, L;
   # Find the base ring.
   GG := List(G);
   FlatG := Flat(GG);
   # Find the operation
   SizeG := Size(FlatG);
   # base = 0 means G is a ring, base > 0 means add mod base.
   base := 0;
   if (Number(FlatG, IsInt) = SizeG) then
      base := Maximum(FlatG) + 1;
   fi;
   L := [];
   if (IsList(H) or IsRing(H)) then
      XX := List(H);
   else
      XX := [H]; # list with 1 element
   fi;
   if (IsList(Y) or IsRing(Y)) then
      YY := List(Y);
   else
      YY := [Y]; # list with 1 element
   fi;
   # assume first that XX and YY are simple lists.
   for i in [1..Size(XX)] do
      for j in [1..Size(YY)] do
         if(IsList(XX[i]) or IsList(YY[j]) ) then
            Add(L, AddC(FlatG, XX[i], YY[j])); # recursively defined for sets of sets
         else
            # find the sum p
            if (base > 0) then
               p := (XX[i] + YY[j]) mod base;
            else
               p := XX[i]+YY[j];
            fi;
            for k in [1..Size(GG)] do
               if (p = GG[k]) then
                  Add(L, GG[k]);
                  break;
               fi;
            od;
         fi;
      od;
   od;
   Sort(L);
   # remove duplicates from the list
   for i in Reversed([2..Size(L)]) do
      for j in Reversed([1..i-1]) do
         if (L[i] = L[j]) then
             Remove(L,i);
             break;
         fi;
      od;
   od;
   return L;
end;

#############################################################################
##
##  MessageToNumber( <string> )
##
##  Converts a single number from the string of characters, with every 
##  letter converting to two digits of the base 10 number.
##
#############################################################################

MessageToNumber := function( s )
   local i, total, strlist, AsciiToCent;
   # Local function converts the decimal form of a character into a number,
   # with Space = 0, A = 1, B = 2, etc., a = 51, b = 52, etc.  
   # Only 94 characters are actually allowed. (those available on a keyboard). 
   AsciiToCent := function( x )
      local r;
      if   x < 33 then r := 0;
      elif x < 34 then r := 47;
      elif x < 48 then r := x+47;
      elif x < 65 then r := x-18;
      elif x < 94 then r := x-64;
      elif x < 127 then r := x-46;
      else r := 0;
      fi;
      return r;
   end;
   strlist := List(s, INT_CHAR);
   total := 0;
   for i in [1..Size(strlist)] do
      total := total * 100 + AsciiToCent(strlist[i]);
   od;
   return total;
end;

#############################################################################
##
##  NumberToMessage( <number> )
## 
##  Reverses the process of MessageToNumber, in that every two digits
##  is converted to a single letter.  0 = Space, 1 = A, 2 = B, etc.
##  95 through 99 are converted to foreign characters, ever though
##  NumberToMessage does not support these foreign characters.
##
#############################################################################

NumberToMessage := function( n )
   local strlist, Temp, m, CentToAscii;
   # Local function converts a number to an ASCII code of a letter.
   CentToAscii := function( x )
      local r;
      if   x < 1  then r := 32;
      elif x < 30 then r := x+64;
      elif x < 47 then r := x+18;
      elif x < 48 then r := 33;
      elif x < 81 then r := x+46;
      elif x < 95 then r := x-47;
      elif x = 95 then r := 196;
      elif x = 96 then r := 203;
      elif x = 97 then r := 214;
      elif x = 98 then r := 220;
      elif x = 99 then r := 223;
      else x := 32;
      fi;
      return r;
   end;
   Temp := n;
   strlist := "";
   while Temp > 0 do
      m := Temp mod 100;
      Temp := QuoInt(Temp,100);
      Add(strlist, CHAR_INT(CentToAscii(m)));
   od;
   strlist := Reversed(strlist);
   return strlist;
end;

#############################################################################
##
##  RtCoset( <group or list of elements>, <subgroup> )
##  
##  Produces a list of right cosets of the subgroup.  If the subgroup is
##  normal, this would be the quotient group.
##
#############################################################################

RtCoset := function(G, H)
   local GG, HH, i, j, k, r, base, p, FlatG, SizeG, t;
   GG := List(G);
   HH := List(H);
   FlatG := Flat(GG); # find the base group
   SizeG := Size(FlatG);
   # base = 0 means G is a group, base > 0 means multiply mod base,
   # base < 0 means add mod base.
   base := 0;
   if (Number(FlatG,IsInt) = SizeG) then
      base := Maximum(FlatG) + 1;
      if (0 in FlatG) then
         base := -base;
      fi;
   fi;
   r := [];
   for i in [1..Size(GG)] do
      if IsList(GG[i]) then
         Add(r, Mult(FlatG, HH, GG[i]));
      else
         t := [];
         for j in [1..Size(HH)] do
            # find the product p
            if (base > 0) then
               p := (HH[j] * GG[i]) mod base;
            elif (base < 0) then
               p := (HH[j] + GG[i]) mod (- base);
            else
               p := HH[j]*GG[i];
            fi;
            for k in [1..Size(GG)] do
               if p = GG[k] then
                  Add(t, GG[k]);
               fi;
            od;
         od;
         Sort(t);
         Add(r, t);
      fi;
   od;
   # remove duplicates from the list
   for i in Reversed([2..Size(r)]) do
      for j in Reversed([1..i-1]) do
         if (r[i] = r[j]) then
             Remove(r,i);
             break;
         fi;
      od;
   od;
   Sort(r);
   return r;
end;

#############################################################################
##
##  LftCoset( <group or list of elements>, <subgroup> )
##  
##  Produces a list of left cosets of the subgroup.  If the subgroup is
##  normal, this would be the quotient group.
##
#############################################################################

LftCoset := function(G, H)
   local GG, HH, i, j, k, r, base, p, FlatG, SizeG, t;
   GG := List(G);
   HH := List(H);
   FlatG := Flat(GG); # find the base group
   SizeG := Size(FlatG);
   # base = 0 means G is a group, base > 0 means multiply mod base,
   # base < 0 means add mod base.
   base := 0;
   if (Number(FlatG,IsInt) = SizeG) then
      base := Maximum(FlatG) + 1;
      if (0 in FlatG) then
         base := -base;
      fi;
   fi;
   r := [];
   for i in [1..Size(GG)] do
      if IsList(GG[i]) then
         Add(r, Mult(FlatG, GG[i], HH));
      else
         t := [];
         for j in [1..Size(HH)] do
            # find the product p
            if (base > 0) then
               p := (GG[i] * HH[j]) mod base;
            elif (base < 0) then
               p := (GG[i] + HH[j]) mod (- base);
            else
               p := GG[i]*HH[j];
            fi;
            for k in [1..Size(GG)] do
               if p = GG[k] then
                  Add(t, GG[k]);
               fi;
            od;
         od;
         Sort(t);
         Add(r, t);
      fi;
   od;
   # remove duplicates from the list
   for i in Reversed([2..Size(r)]) do
      for j in Reversed([1..i-1]) do
         if (r[i] = r[j]) then
             Remove(r,i);
             break;
         fi;
      od;
   od;
   Sort(r);
   return r;
end;

#############################################################################
##
##  Coset( <ring or list of elements>, <subring> )
##
##  For a subring, this produces the cosets using the additive structure.
##  If the subring is an ideal, this produces the quotient ring.
##
#############################################################################

Coset := function(G, H)
   local GG, HH, i, j, k, r, base, p, FlatG, SizeG, t;
   GG := List(G);
   HH := List(H);
   FlatG := Flat(GG); # find the base group
   SizeG := Size(FlatG);
   # base = 0 means G is a ring, base > 0 means add mod base.
   base := 0;
   if (Number(FlatG,IsInt) = SizeG) then
      base := Maximum(FlatG) + 1;
   fi;
   r := [];
   for i in [1..Size(GG)] do
      if IsList(GG[i]) then
         Add(r, AddC(FlatG, GG[i], HH));
      else
         t := [];
         for j in [1..Size(HH)] do
            # find the product p
            if (base > 0) then
               p := (GG[i] + HH[j]) mod base;
            else
               p := GG[i]+HH[j];
            fi;
            for k in [1..Size(GG)] do
               if p = GG[k] then
                  Add(t, GG[k]);
               fi;
            od;
         od;
         Sort(t);
         Add(r, t);
      fi;
   od;
   # remove duplicates from the list
   for i in Reversed([2..Size(r)]) do
      for j in Reversed([1..i-1]) do
         if (r[i] = r[j]) then
             Remove(r,i);
             break;
         fi;
      od;
   od;
   Sort(r);
   return r;
end;

#############################################################################
##
##  P( <list of numbers> )   or   P( [ <list of numbers> ] )
##
##  P is defined to make it easier to enter a Transformation
##
#############################################################################
P := function( arg )
   if 1 = Length(arg) then
      return Transformation(arg[1]);
   else 
      return Transformation(arg);
   fi;
end;

#############################################################################
##
## CycleToPerm converts a product of disjoint cycles to a transformation
##
## PermToCycle converts a transformation, which is a permutation, to a 
##   product of disjoint cycles.
##
#############################################################################

CycleToPerm := AsTransformation;

PermToCycle := function( perm )
   return PermList(ImageListOfTransformation(perm));
end;

#############################################################################
##
##  PermInv( <transformation> )
##
##  If the transformation is invertable, find the inverse as a transformation.
##  Note that the result is unpredictable if the transformation is not invertable.
##
#############################################################################

PermInv := function(x)
   local L, i, u, R;
   if not IsTransformation(x) then
      Print("Input must be a transformation\n");
      return 0;
   fi;
   L := ImageListOfTransformation(x);
   u := Size(L);
   R := List([1..u], x->0);
   for i in [1..u] do
      R[L[i]] := i;
   od;
   return Transformation(R);
end;

#############################################################################
##
##  PermToInt( <permutation in cycles> )
## 
##  Converts a permutation to a positive integer using the natural ordering.
##
#############################################################################

PermToInt := function(x)
   local u, total, i, j;
   total := 1;
   if IsPerm(x) then
      u := LargestMovedPoint(x);
   elif IsTransformation(x) then
      u := DegreeOfTransformation(x);
   else
      Print("Input must be a permutation or transformation\n");
      return 0;
   fi;
   for i in [2..u] do
      for j in [1..i] do
         if (j^x > i^x) then
            total := total + Factorial(i-1);
         fi;
      od;
   od;
   return total;
end;

#############################################################################
##
## NthPerm( <positive integer> )
##
## Finds the nth permutation in their natural ordering.
##
#############################################################################

NthPerm := function(n)
   local u, Temp, S, k, kk, digit, Unfactorial;
   # Local function 
   Unfactorial := function(n)
      # returns the smallest number for which i! >= n
      local index, Temp;
      index := 0;
      Temp := 1;
      while (Temp < n) do
         index := index + 1;
         Temp := Temp * index;
      od;
      return index;
   end;
   if (not IsPosInt(n)) then
      Print("Input must be a positive integer.\n");
      return 0;
   fi;
   u := Unfactorial(n);
   Temp := List([1..u],x->u-x+1);  # u to 1
   S := [ ];
   for k in [1..u] do
      kk := u-k+1;
      digit := QuoInt((n-1) mod Factorial(kk), Factorial(kk-1)) + 1;
      Add(S, Temp[digit]);
      Remove(Temp,digit);
   od;
   S := Reversed(S);
   return PermList(S);
end;   

#############################################################################
##
##  InitQuaternions()
##
##  Defines the quaternion ring / group, using e, i, j, and k as the units.
##  Since students have not learned yet how to enter in rings, this is done
##  as InitTerry:  Set a string to the list of commands, and then Read the 
##  string as input text.
##
#############################################################################

InitQuaternions := function()
local s;
  s := "QQ := QuaternionAlgebra(Rationals);\n";
  s := Concatenation(s,"GenQ := GeneratorsOfAlgebra(QQ);\n");
  s := Concatenation(s,"SizeGenQ := Size(GenQ);\n");
  # If run twice, GenQ will contain more than 4 elements.
  s := Concatenation(s,"e := GenQ[SizeGenQ-3];\n");
  s := Concatenation(s,"i := GenQ[SizeGenQ-2];\n");
  s := Concatenation(s,"j := GenQ[SizeGenQ-1];\n");
  s := Concatenation(s,"k := GenQ[SizeGenQ];\n");
  s := Concatenation(s,"Q := Group(i,j);\n");
  Read(InputTextString(s));
end;

#############################################################################
##
##  InitRing( <string>, [<string>,  ] )
##
##      Sets up defining a ring using the strings as the name of the generators
##
##  DefineRing( <string>, <list of orders>, <matrix of products> )
##
##      Sets up a ring named by the string, using the list of orders for each of 
##      the generators, and a matrix of products of two generators.
##
##  Example:  InitRing("a","b");
##            DefineRing("R",[4,2],[[a,0],[0,b]]);
##   
##      defines a ring R with two generators a and b for which 4*a=0, 2*b=0, 
##      a^2=a, a*b =0, b*a=0, b^2=b.
##
#############################################################################

# This requires the newrings.g package created by Alexander Hulpke
# If it has not been loaded yet, load it first. 
if IsBound(RingByStructureConstants) = false then 
   Read("../gap/newrings.g"); # should be in same place as batch file
fi;

# Global Variables will be set in InitRing
FreeAA := 0; 
GenOfRing := 0;

InitRing := function( arg )
   local i, str, n;
   #  We actually set up a free associative algebra over the list of generators.
   #  this way, the names of the generators can be used when the DefineRing command
   #  is issued. 
   #  We use the trick of putting all of the commands in a string, and then Reading
   #  the string as input text.
   n := Size(arg);
   str := "FreeAA:=FreeAssociativeAlgebra(Rationals";
   for i in [1..n] do
      str := Concatenation(str,",\"",arg[i],"\"");
   od;
   str := Concatenation(str,");\n");
   for i in [1..n] do
      str := Concatenation(str,arg[i],":=FreeAA.",String(i),";\n");
   od;
   str := Concatenation(str,"GenOfRing:=[");
   for i in [1..n] do
      str := Concatenation(str,"\"",arg[i],"\"");
      if i < n then
         str := Concatenation(str,",");
      fi;
   od;
   str := Concatenation(str,"];\n");
   Read(InputTextString(str));
end;

DefineRing := function(Name, CharR, GenTable)   
   local TempI, TempJ, TempK, str, vec, basis, gens, TempT, TempN, 
           TempCk, TempT1, TempVars, TempV;
   # set up vector space to read the GenTable
   gens := GeneratorsOfAlgebra(FreeAA);
   vec := VectorSpace(Rationals,gens);
   basis := Basis(vec);
   # Create a structure constant array from the product matrix.  
   TempT := [];
   TempN := Size(gens);
   for TempI in [1..TempN] do
      TempT1 := [];
      for TempJ in [1..TempN] do
         TempVars := []; # setting up a three dimensional array.
         TempCk := [];
         # work with the I,J entery of the table
         if GenTable[TempI][TempJ] <> 0 then
            TempV := Coefficients(basis,GenTable[TempI][TempJ]);
            for TempK in [1..TempN] do
               if TempV[TempK] <> 0 then    
                  Add(TempVars,TempK); # add this term to the matrix
                  Add(TempCk,TempV[TempK]);
               fi;
            od;
          fi;
          Add(TempT1,[TempVars,TempCk]);
       od;
       Add(TempT,TempT1);
    od;
    Add(TempT,0); # Neither symmetric or anti-symmetric is assumed.
    Add(TempT, 0); # Add zero of the defining ring
    # Here is the trick of setting up a string that will perform the final task,
    # and then reading the string as input text.
    str := Concatenation(Name,":=RingByStructureConstants(",String(CharR),",",
         String(TempT),",",String(GenOfRing),");\n");
    for TempI in [1..TempN] do
       str := Concatenation(str,GenOfRing[TempI],":=",Name,
          ".",String(TempI),";\n");
    od;
    Read(InputTextString(str));
end;
   
#############################################################################
##
##  CheckRing( <possible ring> )
##
##  Currently, it is possible to define a "ring" using RingByStructureConstants
##  that is not really a ring, that is, the distributive or associate laws
##  may fail to hold.  This checks these laws on the generators, since from there
##  one can prove that the laws hold for the whole ring.
##
#############################################################################   

CheckRing := function(R)
   local i,j,k,t,T,L,B;
   L := List(R);
   B := GeneratorsOfRing(R);   
   T := true;
   for i in L do
     for j in L do
        for k in L do
           if ( (i+j)*k <> (i*k) + (j*k)) then
              T := false;
           fi;
        od;
     od;
  od;
  if (not T) then
     Print("Ring is not left distributive.\n");
  else
     for i in L do
        for j in L do
           for k in L do
              if ( i*(j+k) <> (i*j) + (i*k)) then
                 T := false;
              fi;   
           od;
        od;
     od;
     if (not T) then
        Print("Ring is not right distributive.\n");
     else
        for i in L do
           for j in L do
              for k in L do
                 if ( (i*j)*k <> i*(j*k) ) then
                    T := false;
                 fi;
              od;
           od;
        od;
        if T then
            Print("This is a ring.\n");
        else
            Print("Associative law does not hold.\n");
        fi;
     fi;
  fi;
end;    

#############################################################################
##
##  LongDivision( <numerator>, <denominator> )
##
##  Returns [ <quotient>, <remainder> ] for the long division algorithm.
##  Although GAP has similar functions that work for UFD's, this works 
##  for any ring as long as the leading coefficient of the denominator is
##  invertable.  
##
#############################################################################
   
LongDivision := function(f, g)
  local x, F, G, Q, p, i, j, k, m, n, q, r, y;
  if not ( IsUnivariatePolynomial(f) and IsUnivariatePolynomial(g) ) then
     Print("Both polynomials must involve only one variable. \n");
     return fail;
  fi;
  x := IndeterminateOfUnivariateRationalFunction(f);
  if x <> IndeterminateOfUnivariateRationalFunction(g) then
     Print("Both polynomials must be in the same variable. \n");
     return fail;
  fi;
  F := ShallowCopy(CoefficientsOfUnivariatePolynomial(f));
  # F is now mutable
  G := CoefficientsOfUnivariatePolynomial(g);
  j := Size(G);
  k := Size(F) - Size(G); # size of quotient - 1
  Q := [];
  for i in Reversed([0..k]) do
     m := F[i+j] * (G[j]^-1); 
     Add(Q, m);
     # assumes that the leading coefficient of G is invertable.
     for n in [1..j] do
        F[i+n] := F[i+n] - m*G[n];
     od;
  od;
  # now F contains the remainder, and the reverse of Q is the quotient.
  Q := Reversed(Q);
  i := Size(Q);
  q := Sum([1..i],y-> Q[y]*x^(y-1));
  j := Size(F);
  r := Sum([1..j],y-> F[y]*x^(y-1));
  return [q,r];
end;

#############################################################################  
##   
##  SimpleExtension( <algebraic extension of Rationals> )
##
##  Plan : Given A an extention of the Rational numbers, and B an extention of A, etc.
##  find the field C which is a straight extension of the Rationals, by finding a
##  simple extension of Q(a,b,..).
##  Returns a list [C,w,List] for which C is a field Q(w) that is a simple extension 
##  of the Rationals which is isomorphic to the original.  The element w in the new 
##  field corresponds to the second item in the list in the original field. 
##  The primitive elements of the original field are in a list of elements from
##  the new field. 
##
############################################################################# 

#  First, a recursively defined function that will convert an element of the
#  original function into a vector in Rationals.

ToVector := function(w)
   # This converts an element of C to a vector over Q, using recursive 
   # calls to ExtRepOfObj.
   local k ,T, R, n;
   if IsRat(w) then 
      return w;
   else 
      T := ExtRepOfObj(w);
      R := [];
      n := Size(T);
      for k in [1..n] do
         Add(R, ToVector(T[k]));
      od;   
      return Flat(R);
   fi;
end;    
       
SimpleExtension := function(C)
   local A, ExtList, PolyList, PrimList, XList, DegList, TotDeg, 
          P, x, w, p, d, M, i, j, n, L;
   # Given an extension of an extension of a ... of the rationals,
   # this finds a simple extension of the rationals which is equivalent.
   # Since one can only factor polynomials over simple extensions,
   # this will allow us to find the splitting field over any polynomial.
   A := C;
   # first verify that the root field is indeed the rationals
   while A <> LeftActingDomain(A) do
      A := LeftActingDomain(A);
   od;
   if A <> Rationals then
      Print("The root field must be the rationals.\n");
      return fail;
   fi;
   # Next we create arrays for the stages of field extensions.
   A := C;
   ExtList := [];
   PolyList := [];
   PrimList := [];
   XList := [];
   DegList := [];
   TotDeg := 1;
   while A <> Rationals do
      Add(ExtList,A);
      P := DefiningPolynomial(A);
      Add(PolyList,P);
      Add(PrimList,PrimitiveElement(A));
      x := IndeterminateOfUnivariateRationalFunction(P);
      Add(XList,x);
      i := DegreeIndeterminate(P,x);
      Add(DegList,i);
      TotDeg := TotDeg * i;
      A := LeftActingDomain(A);
   od;
   # At this point, the elements in PrimList cannot be added to each other, 
   # unless they are consecutive.  (GAP won't see that one is a member of an
   # extension of an extension of what the other is a member of.)
   # So we convert all into members of the highest field.
   n := Size(ExtList); # number of extension levels
   for i in Reversed([1..(n-1)]) do
      for j in [2..n] do
         PrimList[j] := PrimList[j]*One(PrimList[i]);
      od;
   od;
   # First attempt -- w = sum of the primitive elements (a+b+c)
   w := Sum([1..n], x->PrimList[x]);
   p := 1;
   d := 1;
   repeat 
      M := [];
      for i in [0..(TotDeg-1)] do 
         Add(M, ToVector(w^i));
      od;
      j := Determinant(M);
      # if this is 0, we have to try a different w.
      # try a+b+2c, a+3b+2c, 4a+3b+2c, 4a+3b+5c, 4a+6b+5c, etc.
      # until something works.  (Only a finite number of failures possible.)
      if j = 0 then
         w := w + d*PrimList[p];
         p := p + 1;
         if d < n then 
            d := d + 1;
         fi;
         if p > n then
            p := 1;
         fi;
      fi;
   until j <> 0;
   P := SolutionMat(M, ToVector(-(w^TotDeg)));
   Add(P,1);
   x := XList[n];
   # Use same variable the first field was defined with
   P := UnivariatePolynomial(Rationals,P,x); 
   A := FieldExtension(Rationals,P);
   # Next we want a way to convert an element from the original field to the 
   # new field.
   L := [];
   for i in [1..n] do
      P := SolutionMat(M, ToVector(PrimList[i]));
      Add(L,ObjByExtRep(FamilyObj(PrimitiveElement(A)),P));
   od;
   L := Reversed(L); 
   return [A,w,L];
end;      

#############################################################################
##
##  Factor( <polynomial>, <extension field> )
## 
##  Factors the polynomial over the extension field, putting the factors in
##  this field.  This is actually done by finding a simple extension which
##  is isomorphic to the extension field using SimpleExtension, and then using
##  the GAP Factors command to factor this, and finally converting back to the
##  original field.
##
#############################################################################

RecursiveConvert := function(y,Nlist)
   local i, j, k, var, L, n, T, S, OldN;
   # Takes an element y of a recursive extension of the rationals, and
   # replaces each primitive element with the corresponing element of Nlist.
   # This does not check that the number of levels of y is the same as the length
   # of Nlist -- this should be done at the top level.
   L := ShallowCopy(Nlist); # or else Nlist will change.
   n := Size(L);
   var := L[n];
   T := ExtRepOfObj(y);
   k := Size(T);
   Remove(L);
   if (n = 1) then
      # Last variable, T will return list of rationals.
      S := Sum([1..k],x -> 
      (NumeratorRat(T[x])*(var^(x-1)))/(DenominatorRat(T[x])*One(var)));
   else 
      # T will contain more work
      S := Zero(var);
      for j in [1..k] do
        S := S + RecursiveConvert(T[j],L)*(var^(j-1));
      od;
   fi;
   return S;
end;

ConvertToOriginal := function(PolyList,w,newvar) 
   local i, j, k, n, NewList, V, P, T, m,S;
   # After a Factors command in the simple field produces a mess, we still
   # want to convert this mess back to the extension of the extension...
   # That is, the (a) will be converted to w, and the variable will be 
   # converted to newvar.
   n := Size(PolyList);
   NewList := [];
   for i in [1..n] do
      V := CoefficientsOfUnivariatePolynomial(PolyList[i]);
      k := Size(V);
      P := Zero(w);
      for j in [1..k] do
         T := ExtRepOfObj(V[j]);
         m := Size(T);
         S := Sum([1..m],x -> 
            (NumeratorRat(T[x])*(w^(x-1)))/(DenominatorRat(T[x])*One(w)));
         P := P + S*(newvar^(j-1));
      od;
      Add(NewList,P);
   od;
   return(NewList);
end;

Factor := function(poly, F)
   local Working, done, found, PrimList, A, L, H, h, i, j, k, 
           n, Q, d, x, t, m, s, w, z, P, y;
           
   # if poly is a member of F, assume that we want GAP's original Factors command.
   if (poly in F) then
     return Factors(F, poly);
   fi;         
   # if K is finite, factors the polynomial over K.
   # inefficient, since it uses trial and error. 
   if not ( IsUnivariatePolynomial(poly) ) then
     Print("Polynomial must involve only one variable. \n");
     return fail;
   fi;
   x := IndeterminateOfUnivariateRationalFunction(poly); 
   if Size(F) = infinity then
       if F = Rationals then
          return Factors(poly);
       fi;
       # Check to see if F is an extension of an extension
       A := F;
       while A <> LeftActingDomain(A) do
          A := LeftActingDomain(A);
       od;
       if A <> Rationals then
          Print("The root field must be the rationals.\n");
          return fail;
       fi;
       # Form a list of primitive elements.
       PrimList := [];
       A := F;
       while A <> Rationals do
          Add(PrimList,PrimitiveElement(A));
          A := LeftActingDomain(A);
       od; 
       # if there is only one primitive element, then GAP's Factors command 
       # will work.
       n := Size(PrimList);
       if n = 1 then
          Q := Factors(poly);
          return Q;
       fi;
       # Get all of the primitive elements as members of the highest order field.
       for i in Reversed([1..(n-1)]) do
          for j in [2..n] do
             PrimList[j] := PrimList[j]*One(PrimList[i]);
          od;
       od; 
       PrimList := Reversed(PrimList);
       # First we cheat: we check to see if any of the primitive elements are 
       # roots of the polynomial.
       Q := [];
       Working := poly;
       for i in [1..n] do
          if Value(Working,PrimList[i]) = 0 then
             Add(Q, x-PrimList[i]);
             Working := LongDivision(Working, x-PrimList[i])[1];
          fi;
       od;
       # It is possible at this point that we have a constant.  
       # If so, we will multiply it to the last factor, and quit.
       if DegreeIndeterminate(x*Working,x) = 1 then
          i := Size(Q);
          Q[i] := Q[i] * Working;
          return Q;
       fi;
       # It is also very possible that we will have a first order polynomial.
       # This of course will be prime, so we add it to the list, and quit.
       if DegreeIndeterminate(Working,x) = 1 then
          Add(Q,Working);
          return Q;
       fi;
       # Now comes the hard part: We must first find a simple field extension of F
       # that is isomorphic
       H := SimpleExtension(F);
       # Convert the polynomial to this new field
       A := H[1];
       w := H[2];
       L := H[3];
       z := Indeterminate(A,"z");
       s := CoefficientsOfUnivariatePolynomial(Working);
       t := List(s, i -> RecursiveConvert(i,L));
       k := Size(t);
       Working := Sum([1..k],i-> t[i]*z^(i-1));
       # Now we can use the GAP's factors command
       H := Factors(PolynomialRing(A),Working);
       #Now we still need to convert each of these factors into the old field
       k := Size(H);
       for i in [1..k] do
          s := CoefficientsOfUnivariatePolynomial(H[i]);
          # This will give us a list of elements
          P := Zero(w);
          t := Size(s);
          for j in [1..t] do
             m := ExtRepOfObj(s[j]); # list of rationals
             y := Size(m);
             P := P + Sum([1..y],d -> 
      (NumeratorRat(m[d])*(w^(d-1)))/(DenominatorRat(m[d])*One(w)))*x^(j-1);
          od;
          Add(Q,P);
       od;
       return Q;
   else
       L := List(F);
       n := Size(L);
       Q := [];
       done := false;
       found := false;
       Working := poly;
       while not done do
          d := DegreeOfLaurentPolynomial(Working);
          if d = 1 then
             Add(Q,Working);
             done := true;
          else
             h := QuoInt(d,2); # highest degree polynomial to try
             found := false; # set to true to excape loops
             for k in [1..h] do
                # try first order, then second, etc.
                for m in [0..(n^k-1)] do
                   # Use m as a base n number to create a polynomial of degree k
                   t := One(x)*x^k; # Leading term
                   for s in [0..(k-1)] do
                      t := t + L[(QuoInt(m, n^s) mod n) + 1]*x^s;
                   od;
                   # See if t is a factor.
                   s := LongDivision(Working,t);
                   if s[2] = 0*x then
                      Add(Q,t);
                      Working := s[1];
                      found := true;
                      break; #out of the m loop
                   fi;
                od; # for m
                if found then
                   break; # out of the k loop
                fi;
             od; # for k
             # tried all possible polynomials
             if not found then
                Add(Q,Working); # irreduciable
                done := true;
             fi;
          fi;
       od;  # repeat till all factored
   fi;
   return Q;
end;

#############################################################################
##
##  ViewFactors( <list of factors>, <field>, <list of generator names> )
##
##  Because an element of an extension of an extension is so obtuse, this
##  converts the element to a form that is more readable.  It actually 
##  creates indeterminates which will display as the generators, the way that
##  they should be displayed.  Rational numbers are still displayed with !!.
##
#############################################################################

ViewFactors := function(FactorsList, A, NamesList)
   local i, j, k, XList, NewList, n, V, z, S, L;
   # Redisplay the factors in FactorList, 
   # first define new variables over A
   XList := [];
   for i in NamesList do
      Add(XList,Indeterminate(A,i));
   od;
   L := ShallowCopy(FactorsList);
   if not IsList(L) then
      L := [L];
   fi;
   n := Size(L);
   NewList := [];
   for i in [1..n] do  
      if IsPolynomialFunction(L[i]) then
         z := IndeterminateOfUnivariateRationalFunction(L[i]);
         V := CoefficientsOfUnivariatePolynomial(L[i]);
         k := Size(V);
         S := Sum([1..k], x-> RecursiveConvert(V[x],XList)*(z^(x-1)));
         Add(NewList,S);
      else 
         Add(NewList, RecursiveConvert(L[i],XList));
      fi;
   od;
   if Size(NewList) = 1 then
      return NewList[1];
   else
      return NewList;
   fi;
end;

#############################################################################
##
##  Im( <map>, <element> )
##
##  Works like Image, but this works for the mappings created for extensions
##  of extensions.  
##
#############################################################################

Im := function(map, element)
   local A, L, mapi, PrimList, n, i, j;
   mapi := MappingGeneratorsImages(map);
   A := Source(map);
   # first verify that the root field is indeed the rationals
   while A <> LeftActingDomain(A) do
      A := LeftActingDomain(A);
   od;
   if A <> Rationals then
      Print("The root field must be the rationals.\n");
      return fail;
   fi;
   PrimList := [];
   A := Source(map);
   while A <> Rationals do
      Add(PrimList,PrimitiveElement(A));
      A := LeftActingDomain(A);
   od;
   # At this point, the elements in PrimList cannot be added to each other, 
   # unless they are consecutive.  (GAP won't see that one is a member of an
   # extension of an extension of what the other is a member of.)
   # So we convert all into members of the highest field.
   n := Size(PrimList);
   for i in Reversed([1..(n-1)]) do
      for j in [2..n] do
         PrimList[j] := PrimList[j]*One(PrimList[i]);
      od;
   od;      
   #Next, we find these primitive elements in the first part of mapi,
   #and sort the elements of the second part in the corresponding order.
   L := [];
   for i in [1..n] do
      for j in [1..n] do
         if PrimList[i] = mapi[1][j] then
             Add(L,mapi[2][j]);
             break;
         fi;
      od;
   od;
   L := Reversed(L);
   return RecursiveConvert(element,L);
end;

 
#############################################################################
##
##  CheckHomo( <map>, <list of elements> )
##
##  It is possible to create a homomorphism on an extension of an extension
##  if one does it with the NC option.  This checks whether such a mapping
##  is really a homomorphism by verifying the homomorphism property on the 
##  set of generators. 
##
#############################################################################
     
CheckHomo := function(map, eList)
   local i, j, n, x, test;
   n := Size(eList);
   test := true;
   for i in [1..n] do
      for j in [1..n] do
         if (test = true) and
            (Im(map, eList[i]*eList[j]) <> 
             Im(map,eList[i])*Im(map,eList[j])) then
             Print("f(",eList[i]," * ",eList[j],") is not f(",eList[i],") * f(",eList[j],
                           ")\n");
             test := false;
          fi;
       od;
    od;
    return test;
end;

ChartHomo := function(map, eList, sList)
   local i, j, k, n;
   n := Size(eList);
   if (Size(sList) <> n) then
      Print("Lists must be the same size.");
      return fail;
   fi;
   Print("[ ");
   for i in [1..n] do
      Print(sList[i]," -> ");
      k := 0;
      for j in [1..n] do
         if ( Im(map,eList[i]) = eList[j] ) then 
            k := j; 
         fi;
      od;
      if (k = 0) then 
         Print("?");
      else
         Print(sList[k]);
      fi;
      if (i < n) then
         Print(", ");
      else
         Print(" ]\n");
      fi;
   od;
end;